/* idt entry points */
extern void division_by_zero(void);
extern void illegal_instruction(void);
extern void general_protection(void);
extern void page_fault(void);
extern void event_callback(void);
extern void int_unknown(void);

#if longmode
/* 64bit only */
extern void debug_trap(void);
extern void debug_int3(void);
extern void int_80(void);
#else
/* 32bit only */
extern void xen_hypercall(void);
#endif

#define EMU_MFN(vaddr) (((intptr_t)(vaddr) - (intptr_t)vstart) >> PAGE_SHIFT)
extern char boot[];
extern char vstart[];

extern int grant_frames;
extern struct grant_entry page_aligned grant_table[];

extern uint64_t gdt_mfn[16];
extern uint64_t emu_hcalls[XEN_HCALL_MAX];
extern uint64_t emu_mcalls[XEN_HCALL_MAX];
extern uint64_t emu_faults[XEN_FAULT_MAX];

/* inline asm bits */
static inline ureg_t read_cr0(void)
{
    ureg_t val;
    asm volatile("mov %%cr0,%0\n\t"
		 : "=r" (val));
    return val;
}

static inline ureg_t read_cr2(void)
{
    ureg_t val;
    asm volatile("mov %%cr2,%0\n\t"
		 : "=r" (val));
    return val;
}

static inline ureg_t read_cr3(void)
{
    ureg_t val;
    asm volatile("mov %%cr3,%0\n\t"
		 : "=r" (val));
    return val;
}

static inline ureg_t read_cr4(void)
{
    ureg_t val;
    asm volatile("mov %%cr4,%0\n\t"
		 : "=r" (val));
    return val;
}

static inline void write_cr0(ureg_t val)
{
    asm volatile("mov %%cr0,%0\n\t"
		 :: "r" (val));
}

static inline void write_cr3(ureg_t cr3)
{
    asm volatile("mov %0, %%cr3;\n"
		 :: "r" (cr3)
		 : "memory");
}

static inline void flush_tlb(void)
{
    ureg_t tmpreg;

    asm volatile("mov %%cr3, %0;              \n"
		 "mov %0, %%cr3;  # flush TLB \n"
		 : "=r" (tmpreg)
		 :: "memory");
}

static inline void flush_tlb_addr(ureg_t va)
{
    asm volatile("invlpg (%0)" ::"r" (va) : "memory");
}

static inline void rdmsr(uint32_t msr, uint32_t *ax, uint32_t *dx)
{
    asm volatile("rdmsr"
		 : "=a" (*ax), "=d" (*dx)
		 : "c" (msr));
}

static inline void wrmsr(uint32_t msr, uint32_t ax, uint32_t dx)
{
    asm volatile("wrmsr"
		 : /* no outputs */
		 : "c" (msr), "a" (ax), "d" (dx));
}

static inline void wrmsrl(uint32_t msr, uint64_t val)
{
    uint32_t ax = (uint32_t)val;
    uint32_t dx = (uint32_t)(val >> 32);
    wrmsr(msr, ax, dx);
}
